/*
	System: Structured text retrieval tool sgrep.
	Module: eval.c
	Author: Pekka Kilpelinen & Jani Jaakkola
	Description: Handles the evaluation of sgrep expressions, thus
		     implementing the actual semantics of sgrep language.
		     used through eval() function
	Version history: Original version February 1995 by JJ & PK
	Copyright: University of Helsinki, Dept. of Compter Science
		   Distributed under GNU General Public Lisence
		   See file COPYING for details
*/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "sgrep.h"

#if defined(OPTIMIZE_SORTS)
/* Define this if you want always test that nest optimization works */
#define ASSERT_NESTS
#endif
    
/*
 * Sometimes regions need to be stacked or queued
 */
#define DEFAULT_STACK_SIZE 1024

typedef struct {
    struct SgrepStruct *sgrep;
    struct REGION *tmp_stack;
    int tmp_stack_size;
} Evaluator;

struct GC_LIST *recursive_eval(Evaluator *,struct TREE_NODE *root);
struct GC_LIST *eval_operator(Evaluator *,struct TREE_NODE *root);
struct GC_LIST *or(struct GC_LIST *,struct GC_LIST *);
struct GC_LIST *nest_order(Evaluator *, struct GC_LIST *,struct GC_LIST *,int);
struct GC_LIST *quote(struct GC_LIST *,struct GC_LIST *,int);
struct GC_LIST *in(struct GC_LIST *,struct GC_LIST *,int);
struct GC_LIST *containing(Evaluator *,struct GC_LIST *,struct GC_LIST *,int);
struct GC_LIST *extracting(struct GC_LIST *,struct GC_LIST *);
struct GC_LIST *outer(struct GC_LIST *);
struct GC_LIST *inner(Evaluator*, struct GC_LIST *);
struct GC_LIST *concat(struct GC_LIST *);
struct GC_LIST *join(struct GC_LIST *,int number);
struct GC_LIST *equal(struct GC_LIST *,struct GC_LIST *,int);
RegionList *parenting(Evaluator *,RegionList *l, RegionList *r);

int free_tree_node(struct TREE_NODE *node);

struct GC_LIST *eval(struct SgrepStruct *sgrep,struct TREE_NODE *root) {
    struct GC_LIST *r;
    Evaluator evaluator;

    evaluator.sgrep=sgrep;
    evaluator.tmp_stack_size=DEFAULT_STACK_SIZE;
    evaluator.tmp_stack=(struct REGION *)sgrep_malloc(DEFAULT_STACK_SIZE*
	sizeof(struct REGION));
    r=recursive_eval(&evaluator,root);
    if (evaluator.tmp_stack) sgrep_free(evaluator.tmp_stack);
    return r;
}

/*
 * Recursively evaluates parse tree using operation functions
 * root points the root node of parse tree
 */
struct GC_LIST *recursive_eval(Evaluator *evaluator,struct TREE_NODE *root)
{	
#ifdef DEBUG
	static int depth=0;
	int i;
#endif
	struct GC_LIST *a;
	SGREPDATA(evaluator);
	a=root->GC_list;

#ifdef DEBUG
	depth++;
	for(i=0;i<depth;i++) fputc(' ',stderr);
	fprintf(stderr,"Evaluating oper %s l_label=%d r_label=%d\n",
		give_oper_name(root->oper),root->label_left,root->label_right);
#endif
	assert(root->oper!=INVALID);

	/* If this is a leaf node, we just use leafs gc list */
	if ( a==NULL && root->oper==PHRASE )
	{
	        if (sgrep->index_file && root->leaf->GC_list==NULL) {
		    assert(root->leaf->phrase!=NULL);
		    assert(sgrep->index_reader);
		    root->leaf->GC_list=index_lookup(
			sgrep->index_reader,
			root->leaf->phrase->s);
		}
		assert(root->leaf->GC_list!=NULL);

		a=root->leaf->GC_list;
		root->leaf->GC_list=NULL;
		a->refcount=root->refcount;
#ifdef DEBUG
		for(i=0;i<depth;i++) fputc(' ',stderr);
		fprintf(stderr,"Using phrase list %s\n",root->leaf->phrase->s);
#endif
	}
	/* If gc_list is still NULL, it means that it hasn't been evaluated yet */
	if ( a==NULL )
	{
#ifdef DEBUG
		for(i=0;i<depth;i++) fputc(' ',stderr);
		fprintf(stderr,"Calling real eval oper %s\n",
			give_oper_name(root->oper));
#endif
		a=eval_operator(evaluator,root);
		a->refcount=root->refcount;
		/* We free subtrees unneeded gclists */
		if (free_tree_node(root->left))
		{
#ifdef DEBUG
			for(i=0;i<depth;i++) putc(' ',stderr);
			fprintf(stderr,"label %d freed (left)\n",root->label_left);
#endif
		}
		if (free_tree_node(root->right))
		{
#ifdef DEBUG
			for(i=0;i<depth;i++) putc(' ',stderr);
			fprintf(stderr,"label %d freed (right)\n",root->label_right);
#endif
		}
	}
#ifdef DEBUG
	else
	{
		for(i=0;i<depth;i++) fputc(' ',stderr);
		fprintf(stderr,"Using already known list\n");
	}
#endif		
	
	/* Keeps track of longest used gc list */
	if (LIST_SIZE(a)>stats.longest_list)
		stats.longest_list=LIST_SIZE(a);
#ifdef ASSERT_NESTS
	/* We check that if list isn't marked as nested, it really isn't */
	if (!a->nested)
	{
		struct REGION reg1,reg2;
		struct GC_POINTER p;
		
		start_region_search(a,&p);
		get_region(&p,&reg1);
		get_region(&p,&reg2);
		while (reg2.start!=-1)
		{
			assert(reg1.end<reg2.end);
			reg1=reg2;
			get_region(&p,&reg2);
		}
	}
#endif	

	root->GC_list=a;	

#ifdef DEBUG
	for(i=0;i<depth;i++) fputc(' ',stderr);
	fprintf(stderr,"eval done\n");
	depth--;
#endif
	return a;
}

/*
 * Handles the actual evaluation of some operation
 */
struct GC_LIST *eval_operator(Evaluator *evaluator,struct TREE_NODE *root)
{
	struct GC_LIST *a,*l,*r;

	a=NULL;
	assert(root->left!=NULL);

	
	/* Let's evaluate left and right subtrees first */
	l=recursive_eval(evaluator,root->left);
	/* if root->right==NULL this is a function, and it won't have 
	   right sub tree */
	if (root->right==NULL) r=NULL;
	else r=recursive_eval(evaluator,root->right);
	
	/* Let's call the operation function */
	switch (root->oper) {
	case OR:
		a=or(l,r);
		break;
	case ORDERED:
	case L_ORDERED:
	case R_ORDERED:
	case LR_ORDERED:
		a=nest_order(evaluator,l,r,root->oper);
		break;
	case QUOTE:
	case L_QUOTE:
	case R_QUOTE:
	case LR_QUOTE:
		a=quote(l,r,root->oper);
		break;
	case IN:
		a=in(l,r,FALSE);
		break;
	case NOT_IN:
		a=in(l,r,TRUE);
		break;
	case CONTAINING:
		a=containing(evaluator,l,r,FALSE);
		break;
	case NOT_CONTAINING:
		a=containing(evaluator,l,r,TRUE);
		break;
/* Start PK Febr 95 */
	case EQUAL:
		a=equal(l,r,FALSE);
		break;
	case NOT_EQUAL:
		a=equal(l,r,TRUE);
		break;
/* End PK Febr 95 */
	case PARENTING:
	        a=parenting(evaluator,l,r);
		break;
	case OUTER:
		a=outer(l);
		break;
	case INNER:
		a=inner(evaluator,l);
		break;
	case EXTRACTING:
		a=extracting(l,r);
		break;
	case CONCAT:
		a=concat(l);
		break;
	case JOIN:
		a=join(l,root->number);
		break;
	default:
	       sgrep_error(evaluator->sgrep, 
			   "Unknown operator in parse tree (%d)\n",
			   root->oper);
	       assert(0 && "Unknown operator in parse tree");
	       break;
	}
	
	return a;
}

/*
 * Decrements tree nodes reference counter, and frees nodes gc list if
 * counter comes down to 0. Returns 1 if something was freed, 0
 * otherwise
 */
int free_tree_node(struct TREE_NODE *node)
{
	if (node==NULL) return 0; /* This was a leaf or function node */ 
	if (node->GC_list!=NULL) 
	{
		if (node->GC_list->refcount!=-1)
			node->GC_list->refcount--;
		assert(node->GC_list->refcount>=0 
			|| node->GC_list->refcount==-1);
		if (node->GC_list->refcount==0)
		{
			free_gclist(node->GC_list);
			node->GC_list=NULL;
			return 1;
		}
	}
	return 0;
}

#ifdef PROGRESS_REPORTS
/*
 * Shows a progress report on stderr
 */
void report_progress(char *line,int size, int now)
{
	fprintf(stderr,"%s %d%% done%s\r",
		line,(100*now)/size,"                  ");
	fflush(stderr);
}
#endif

/*
 * Gives first region from two gc_lists, eliminating same regions.
 */
struct REGION first_of(struct GC_POINTER *lp,struct GC_POINTER *rp)
{
	struct REGION l_reg,r_reg;
	
	/* quite straightforward limiting of two gc lists.
	   same regions are concatanated */	
	get_region(lp,&l_reg);
	get_region(rp,&r_reg);
	if (r_reg.start!=-1 && l_reg.start!=-1)
	{
		if (l_reg.start<r_reg.start)
		{
			prev_region(rp,&r_reg);
			return l_reg;
		} else if (l_reg.start>r_reg.start)
		{
			prev_region(lp,&l_reg);
			return r_reg;
		} else if (l_reg.end<r_reg.end)
		{
			prev_region(rp,&r_reg);
			return l_reg;
		} else if (l_reg.end>r_reg.end)
		{
			prev_region(lp,&l_reg);
			return r_reg;
		} else 
		{
			return r_reg;
		}
	}
	if (r_reg.start!=-1) return r_reg;
	if (l_reg.start!=-1) return l_reg;
	/* Both lists were empty, we return (-1,-1) */
	return r_reg;
}

/*
 * Handles or operation
 */
struct GC_LIST *or(struct GC_LIST *l,struct GC_LIST *r)
{
	struct GC_POINTER lp,rp;
	struct GC_LIST *a;
	struct REGION tmp;
#ifdef OPTIMIZE_SORTS
	struct REGION prev;
#endif	
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif
	SGREPDATA(l);

#ifdef DEBUG
	fprintf(stderr,"or called\n");
#endif
	stats.or_oper++;
	a=new_region_list(sgrep);
#ifdef OPTIMIZE_SORTS
	prev.start=-1;
	prev.end=-1;
#endif
	start_region_search(l,&lp);
	start_region_search(r,&rp);

#ifdef PROGRESS_REPORTS
	prog_start=LIST_SIZE(r)+LIST_SIZE(l);
	prog=0;
	oper_name="or";
#endif

	for(tmp=first_of(&lp,&rp);tmp.start!=-1;tmp=first_of(&lp,&rp))
	{
#ifdef OPTIMIZE_SORTS
		if ( tmp.end<=prev.end )
		{
			/* We had nesting */
			a->nested=TRUE;
		}
#endif
		add_region(a,tmp.start,tmp.end);
#ifdef OPTIMIZE_SORTS
		prev=tmp;
#endif
	}
	return a;
}
	
/* 
 * Handles ordering which produces possibly nesting gc-lists. 
 */
struct GC_LIST *nest_order(Evaluator *evaluator, 
			   struct GC_LIST *l,struct GC_LIST *r,int type)
{
    struct GC_POINTER lp,rp;
    struct GC_LIST *a;
    struct REGION r_reg,l_reg;              
    int nest_depth=0;
    int nestings;
    int s,e;
    SGREPDATA(evaluator);
#ifdef PROGRESS_REPORTS
    char *oper_name;
    int prog;int prog_start;
#endif

#ifdef DEBUG
    fprintf(stderr,"nest_order called\n");
#endif
    start_region_search(r,&rp);
    
    stats.order++;
    a=new_region_list(sgrep);
    a->nested=l->nested || r->nested;
#ifdef DEBUG
    if (a->nested) fprintf(stderr,"inherited nesting\n");
#endif

#ifdef PROGRESS_REPORTS
    prog_start=LIST_SIZE(r);
    prog=0;
    switch (type) {
    case L_ORDERED:
	oper_name="_.";
	break;
    case R_ORDERED:
	oper_name="._";
	break;
    case LR_ORDERED:
	oper_name="__";
	break;
    default:
	oper_name="..";
    }
#endif
    nestings=0;
    start_end_sorted_search(l,&lp);
    get_region(&lp,&l_reg);
    get_region(&rp,&r_reg);
    /* If left or right region list was empty, we can return empty list */
    if (l_reg.start==-1 || r_reg.start==-1) return a;
    
    do
    {
	if (l_reg.end<r_reg.start && l_reg.start!=-1 )
	{
	    /* left region is first. Add to nest_stack
	       and nest queue */
	    if (nest_depth==evaluator->tmp_stack_size)
	    {
		evaluator->tmp_stack_size+=evaluator->tmp_stack_size/2;
		evaluator->tmp_stack=(struct REGION *)sgrep_realloc(
		    evaluator->tmp_stack,
		    evaluator->tmp_stack_size*sizeof(struct REGION));
	    }
	    evaluator->tmp_stack[nest_depth++]=l_reg;
	    nestings=0;
#ifdef DEBUG
	    if (nest_depth==1)
		fprintf(stderr," New q");
	    else fprintf(stderr," +");
	    fprintf(stderr,"(%d:%d)",l_reg.start,l_reg.end);
#endif
	    get_region(&lp,&l_reg);
	}
	else if (nest_depth>0)
	{
#ifdef DEBUG
	    fprintf(stderr," %d",r_reg.end);
#endif
	    if (type==L_ORDERED || type==LR_ORDERED)
		s=evaluator->tmp_stack[--nest_depth].end+1;
	    else s=evaluator->tmp_stack[--nest_depth].start;
	    if (type==R_ORDERED || type==LR_ORDERED)
		e=r_reg.start-1;
	    else e=r_reg.end;
	    
	    if (e>=s)
	    {
		/* If we have taken region from nest stack
		 * twice in row, it probably means, that
		 * we have a nested result list */
		nestings++;
		if (nestings==2)
		{
#ifdef DEBUG
		    if (!a->nested)
			fprintf(stderr,"nesting order detecded\n");
#endif
		    a->nested=TRUE;
		    list_set_sorted(a,NOT_SORTED);
		}
		add_region(a,s,e);
	    }			
	    get_region(&rp,&r_reg);
	} else 
	{
	    get_region(&rp,&r_reg);
	}
    } while ( r_reg.start!=-1 );
    return a;
}

/*
 * Handles in operation 
 */
struct GC_LIST *in(struct GC_LIST *l,struct GC_LIST *r, int not)
/* Changed by PK in Febr 95 to capture the semantics of _proper_
containment */
{	
	struct GC_POINTER lp,rp;
	struct GC_LIST *a,*r2;
	struct REGION r_reg,l_reg,r_reg2;
	char *oper_name;
#ifdef PROGRESS_REPORTS
	int prog;int prog_start;
#endif
	SGREPDATA(l);

#ifdef DEBUG
	fprintf(stderr,"in called\n");
#endif
	if (not) 
	{
		stats.not_in++;
		oper_name="not in";
	} else 
	{
		stats.in++;
		oper_name="in";
	}
	a=new_region_list(sgrep);
	
#ifdef OPTIMIZE_SORTS
	a->nested=l->nested;
#endif
	
	start_region_search(l,&lp);
	get_region(&lp,&l_reg);
	
	
       /* 
 	* To simplify things we do an outer function on right gc_list 
 	*/
#ifdef OPTIMIZE_SORTS
	if (r->nested)
	{
#endif
		r2=outer(r);
		r=r2;
#ifdef OPTIMIZE_SORTS
	} else r2=NULL;
#endif
	start_region_search(r,&rp);

#ifdef PROGRESS_REPORTS
	prog_start=LIST_SIZE(l)+LIST_SIZE(r);
	prog=0;
#endif

	get_region(&rp,&r_reg);
	while (r_reg.start!=-1 && l_reg.start!=-1)
	{
#ifdef DEBUG
		fprintf(stderr,"in: left=(%d,%d) right=(%d,%d)\n",
			l_reg.start,l_reg.end,
			r_reg.start,r_reg.end);
#endif
		if (l_reg.start<r_reg.start)
		{
			/* Left region starts before right -> can't be
			   in right region or any right region that follows
			   current one */
			if (not) add_region(a,l_reg.start,l_reg.end);
			get_region(&lp,&l_reg);
		} else /* l_reg.start>=r_reg.start */
		{
			if (l_reg.end<=r_reg.end)
			{
				/* left region is in right region */
/* Start PK Febr 95 */
				if (l_reg.start>r_reg.start ||
				    l_reg.end<r_reg.end)
				{	/* inclusion is proper */
					if (!not) add_region(a,l_reg.start,l_reg.end);
					get_region(&lp,&l_reg);
				} else { /* l_reg == r_reg */ 
					if (not) add_region(a,l_reg.start,l_reg.end);
					get_region(&lp,&l_reg);
				}
/* End PK Febr 95 */
			} else 	if (l_reg.start==r_reg.start)
			{
				/* Regions start from same place. Because
				no right region after current one can start 
				from same place we can skip left region */
				if (not) add_region(a,l_reg.start,l_reg.end);
				get_region(&lp,&l_reg);
			} else
			{
				/* left and right region are overlapping */
#ifdef DEBUG
				fprintf(stderr,"in overlap\n");
#endif
				get_region(&rp,&r_reg2);
				if (r_reg2.start==-1)
				{
					/* All right regions have been scanned */
					if ( l_reg.start > r_reg.end )
					{
						/* Left region end after last right region. 
						   We can fall out of loop */
						r_reg=r_reg2;
					} else
					{
						/* Next left region might still be in right 
						   region */
						if (not) add_region(a,l_reg.start, l_reg.end);
						get_region(&lp,&l_reg);
					}
				} else
				{
					/* There are still right regions */
					if ( l_reg.start >= r_reg2.start )
					{
						/* Since left region starts after new right region,
						 * We can safely skip previous right */
						r_reg=r_reg2;
					} else
					{
						/* Left region is not in previous or next 
						   right region */
						prev_region(&rp,&r_reg2);
						if (not) add_region(a,l_reg.start, l_reg.end);
						get_region(&lp,&l_reg);
					}
				}
			}
		}
	}

#ifdef DEBUG
	fprintf(stderr,"in fall out\n");
#endif
/* If we have "not in" and right gc_list is empty, we need to copy
   rest of left list */
	if (not)
	{
		while (l_reg.start!=-1)
		{
			add_region(a,l_reg.start,l_reg.end);
			get_region(&lp,&l_reg);
		}
	}

/* because we created list r2 here, we free it here */
	if (r2!=NULL) free_gclist(r2);
	return a;
}

/* 
 * Handles outer function 
 */
struct GC_LIST *outer(struct GC_LIST *gcl)
{
	struct GC_POINTER p;
	struct REGION reg1,reg2;
	struct GC_LIST *a;
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif
	SGREPDATA(gcl);
#ifdef PROGRESS_REPORTS
	oper_name="outer";
	prog_start=LIST_SIZE(gcl);
	prog=0;
#endif

	stats.outer++;
	reg2.start=0;
	a=new_region_list(sgrep);
	start_region_search(gcl,&p);
	get_region(&p,&reg1);

	/* if we had empty gc list */
	if (reg1.start==-1) return a;
		
	/* If there are many regions starting from same place, we choose the
	   longest */
	get_region(&p,&reg2);
	while (reg2.start==reg1.start && reg2.end>reg1.end)
	{
		reg1=reg2;
		get_region(&p,&reg2);
	}
	
	while(reg1.start!=-1 && reg2.start!=-1)
	{
		if (reg2.end>reg1.end && reg2.start!=reg1.start)
		{
			/* reg2 ends after reg1 -> no nesting */
			add_region(a,reg1.start,reg1.end);
			reg1=reg2;
		}
		get_region(&p,&reg2);
		/* If regions start from same place, nesting is guaranteed */
		if (reg2.start==reg1.start)
		{
			reg1=reg2;
			get_region(&p,&reg2);
		}
	}
	add_region(a,reg1.start,reg1.end);
	return a;
}

/*
 * Handles inner function 
 */
struct GC_LIST *inner(Evaluator *evaluator,struct GC_LIST *gcl)
{
    struct GC_POINTER p;
    int inq_ind=0;
    struct GC_LIST *a=NULL;
    struct REGION n_reg,c_reg;
    int i;
    struct REGION *inner_stack;
    SGREPDATA(evaluator);
#ifdef PROGRESS_REPORTS
    char *oper_name;
    int prog;int prog_start;
#endif

#ifdef DEBUG
    fprintf(stderr,"inner called\n");
#endif	
    stats.inner++;
    a=new_region_list(sgrep);
    inner_stack=evaluator->tmp_stack;

#ifdef PROGRESS_REPORTS
    prog_start=LIST_SIZE(gcl);
    prog=0;
    oper_name="inner";
#endif

    start_region_search(gcl,&p);
    get_region(&p,&c_reg);
    while (c_reg.start!=-1) {
	get_region(&p,&n_reg);
	assert(n_reg.start>=c_reg.start || n_reg.start==-1 );
	if ( n_reg.start>c_reg.end || n_reg.start==-1 )
	{
	    /* n_reg and c_reg are separate. Therefore c_reg must
	       be innermost */
	    /* Now we can empty inner_stack */
#ifdef DEBUG
	    fprintf(stderr,"empty inner stack (%d regions)\n",inq_ind);
#endif
	    for (i=0;i<inq_ind;i++)
	    {
		assert(inner_stack[i].start<=c_reg.start);

		if (inner_stack[i].end<c_reg.end)
				/* Region in inner_stack was innermost */
		    add_region(a,inner_stack[i].start,
			       inner_stack[i].end);
	    }
	    inq_ind=0;
	    add_region(a,c_reg.start,c_reg.end);
	} else if ( n_reg.end>c_reg.end )
	{
	    /* n_reg and c_reg are overlapping. Let's add c_reg
	       to inner_stack */
	    if (evaluator->tmp_stack_size==inq_ind)
	    {
		evaluator->tmp_stack_size+=evaluator->tmp_stack_size/2;
		inner_stack=(struct REGION *)
		    sgrep_realloc(inner_stack,
				  evaluator->tmp_stack_size*
				  sizeof(struct REGION));
		evaluator->tmp_stack=inner_stack;
	    }
	    inner_stack[inq_ind++]=c_reg;
	} else {
	    /* if neither of the previous if's was taken, 
	       c_reg contains n_reg. We remove regions containing n_reg from
	       inner_stack */
	    while(inq_ind &&  
		  n_reg.start>=inner_stack[inq_ind-1].start &&
		  n_reg.end<=inner_stack[inq_ind-1].end )
	    {
		inq_ind--;
	    }
	}
	c_reg=n_reg;
	if (inq_ind) 
	    assert(c_reg.start<inner_stack[inq_ind-1].start ||
		   c_reg.end>inner_stack[inq_ind-1].end);
    }
    return a;
}

struct GC_LIST *containing(Evaluator *evaluator,struct GC_LIST *l,
			   struct GC_LIST *r,int not)
/* Changed by PK in Febr 95 to capture the semantics of _proper_
containment */
{
	struct GC_POINTER lp,rp;
	struct GC_LIST *a,*r2;
	struct REGION r_reg,l_reg;
	SGREPDATA(evaluator);
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif

#ifdef DEBUG
	fprintf(stderr,"containing called\n");
#endif
	if (not) stats.not_containing++; else stats.containing++;
	a=new_region_list(sgrep);
#ifdef OPTIMIZE_SORTS
	a->nested=l->nested;
#endif
	start_region_search(l,&lp);
	get_region(&lp,&l_reg);

/* To simplify things we do an inner function on right gc_list */
#ifdef OPTIMIZE_SORTS
	if (r->nested)
	{
#endif
		r2=inner(evaluator,r);
		r=r2;
#ifdef OPTIMIZE_SORTS
	} else r2=NULL;
#endif

#ifdef PROGRESS_REPORTS
	oper_name= (not) ? "not containing" : "containing";
	prog=0;
	prog_start=LIST_SIZE(l)+LIST_SIZE(r);
#endif
	start_region_search(r,&rp);
	
	get_region(&rp,&r_reg);
	while (r_reg.start!=-1 && l_reg.start!=-1)
	{
		if ( l_reg.start>r_reg.start )
		{
			/* right starts before left */
			get_region(&rp,&r_reg);
		} else if ( l_reg.end>=r_reg.end )
		{
			/* left contains right */
/* Start PK Febr 95 */
			if (l_reg.start<r_reg.start ||
			    l_reg.end>r_reg.end)
			{	/* Containment is proper */
				if (!not) add_region(a,l_reg.start,l_reg.end);
				get_region(&lp,&l_reg);
			} else { /* l_reg == r_reg */
				if (not) add_region(a,l_reg.start,l_reg.end);
				get_region(&lp,&l_reg);
			}
/* End PK Febr 95 */
		} else {
			/* left comes after right */
			if (not) add_region(a,l_reg.start,l_reg.end);
			get_region(&lp,&l_reg);
		}	
	}
	/* When right list ended, there still might be something in left list */
	while (not && l_reg.start!=-1)
	{
		add_region(a,l_reg.start,l_reg.end);
		get_region(&lp,&l_reg);
	}
/* because we created list r2 here, we free it here */
	if (r2!=NULL) free_gclist(r2);
	return a;
}

struct GC_LIST *equal(struct GC_LIST *l,struct GC_LIST *r,int not)
/* Intersection of GC_LISTs *l and *r */
/* PK Febr '95 */
{
	struct GC_POINTER lp,rp;
	struct GC_LIST *a;
	struct REGION r_reg,l_reg;
	SGREPDATA(l);
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif

#ifdef DEBUG
	fprintf(stderr,"equal called\n");
#endif
	if (not) stats.not_equal++; else stats.equal++;
	a=new_region_list(sgrep);
#ifdef OPTIMIZE_SORTS
	a->nested=l->nested;
#endif
	start_region_search(l,&lp);
	get_region(&lp,&l_reg);


#ifdef PROGRESS_REPORTS
	oper_name= (not) ? "not equal" : "equal";
	prog=0;
	prog_start=LIST_SIZE(l)+LIST_SIZE(r);
#endif
	start_region_search(r,&rp);
	get_region(&rp,&r_reg);


	while (r_reg.start!=-1 && l_reg.start!=-1)
	{
		if ( l_reg.start<r_reg.start )
		{
			if (not) add_region(a,l_reg.start,l_reg.end);
			get_region(&lp,&l_reg);
		} else if ( r_reg.start<l_reg.start )
		{
			get_region(&rp,&r_reg);
		} else  /*  r_reg.start=l_reg.start */
			if ( l_reg.end<r_reg.end )
			{
				if (not) add_region(a,l_reg.start,l_reg.end);
				get_region(&lp,&l_reg);
			} else if ( r_reg.end<l_reg.end )
			{
				get_region(&rp,&r_reg);
			} else /* l_reg = r_reg */
			{
				if (!not) add_region(a,l_reg.start,l_reg.end);
				get_region(&rp,&r_reg);
				get_region(&lp,&l_reg);
			}
	}
	/* When right list ended, there still might be something in left list */
	while (not && l_reg.start!=-1)
	{
		add_region(a,l_reg.start,l_reg.end);
		get_region(&lp,&l_reg);
	}
	return a;

} /* END equal(struct GC_LIST *l,struct GC_LIST *r,int not) */

/*
 * Here we implement concat operation, which concats all overlapping regions
 * into one region and removes all nestings
 */
struct GC_LIST *concat(struct GC_LIST *l)
{
    SGREPDATA(l);
	struct GC_POINTER lp;
	struct GC_LIST *a;
	struct REGION reg1,reg2;
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif

#ifdef DEBUG
	fprintf(stderr,"concat called\n");
#endif
	stats.concat++;
	a=new_region_list(sgrep);
	start_region_search(l,&lp);
	get_region(&lp,&reg1);
	
	/* We had empty list */
	if (reg1.start==-1) return a;

#ifdef PROGRESS_REPORTS
	oper_name="concat";
	prog=0;
	prog_start=LIST_SIZE(l);
#endif
	get_region(&lp,&reg2);
	
	while (reg2.start!=-1)
	{
		if (reg2.start>reg1.end+1)
		{
			/* separate regions, no concat */
			add_region(a,reg1.start,reg1.end);
			reg1=reg2;
		} else if ( reg2.end>reg1.end )
		{
			/* We found overlapping */
			reg1.end=reg2.end;
		}
		get_region(&lp,&reg2);
	}
	add_region(a,reg1.start,reg1.end);
	return a;
}

/*
 * Here we implement extracting operation
 */
struct GC_LIST *extracting(struct GC_LIST *l,struct GC_LIST *r)
{
    SGREPDATA(l);
	struct GC_POINTER lp,rp,tmpp;
	struct GC_LIST *a,*r2,*tmp,*new_tmp;
	struct REGION l_reg,r_reg;
	int prev_s=-1;
	int prev_e=-1;
	int last_tmp;
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif

#ifdef DEBUG
	fprintf(stderr,"extracting called\n");
#endif
	stats.extracting++;
	/* to simplify things we do concat on right gc_list. Result stays
	   the same anyway */
	r2=concat(r);
	r=r2;
	
	a=new_region_list(sgrep);
#ifdef OPTIMIZE_SORTS
	a->nested=l->nested;
#endif
	tmp=new_region_list(sgrep);
	start_region_search(tmp,&tmpp);

#ifdef PROGRESS_REPORTS
	oper_name="extracting";
	prog=0;
	prog_start=LIST_SIZE(l);
#endif
	
	start_region_search(l,&lp);
	get_region(&lp,&l_reg);
	start_region_search(r,&rp);
	get_region(&rp,&r_reg);

	while ( l_reg.start!=-1 )
	{
		if ( l_reg.end<r_reg.start || r_reg.start==-1 )
		{
			/* Regions are separate, left starting first.
 			   no cutting */
 			if ( prev_s!=l_reg.start || prev_e!=l_reg.end )
 			{				
#ifdef DEBUG		
	fprintf(stderr,"extracting adding 1(%d,%d)\n",l_reg.start,l_reg.end);
#endif
				prev_s=l_reg.start;
				prev_e=l_reg.end;
				add_region(a,l_reg.start,l_reg.end);
			}
			l_reg=first_of(&lp,&tmpp);
		} else if ( r_reg.end<l_reg.start )
		{
			/* Regions are separate right starting first.
			   we skip right */
#ifdef DEBUG
			fprintf(stderr,"skipping right, left=(%d,%d) right=(%d,%d)\n",
				l_reg.start,l_reg.end,r_reg.start,r_reg.end);
#endif
			get_region(&rp,&r_reg);
		} else
		{
			/* We need to do clipping. */
			new_tmp=new_region_list(sgrep);
/* OBSOLETE			new_tmp->start_sorted_list=NULL;
   must_be_sorted=FALSE; */
			last_tmp=-1;
#ifdef DEBUG
			fprintf(stderr,"cutting loop, cutter (%d,%d)\n",r_reg.start,r_reg.end);
#endif
			while ( l_reg.start!=-1 && l_reg.start<=r_reg.end )
			{
			    if (l_reg.start<r_reg.start && 
				( prev_s!=l_reg.start ||
				  prev_e!=r_reg.start-1 ))
			    {
				prev_s=l_reg.start;
				prev_e=r_reg.start-1;
				
#ifdef DEBUG
				fprintf(stderr,"extracting adding 2(%d,%d)\n",l_reg.start,r_reg.start-1);
#endif		
				add_region(a,l_reg.start,r_reg.start-1);
			    }
			    if (r_reg.end<l_reg.end)
			    {
#ifdef DEBUG
				fprintf(stderr,"(%d,%d)<-new_tmp\n",r_reg.end+1,l_reg.end);
#endif
				if (l_reg.end<last_tmp) {
				    list_set_sorted(new_tmp,NOT_SORTED);
				}
				add_region(new_tmp,r_reg.end+1,l_reg.end);
				last_tmp=l_reg.end;
			    }
			    l_reg=first_of(&lp,&tmpp);
			}
			if (l_reg.start!=-1) prev_region(&lp,&l_reg);
				
			assert (tmpp.ind==tmp->length &&
				tmpp.node->next==NULL);

			free_gclist(tmp);
#ifdef PROGRESS_REPORTS
			prog--;
			prog_start+=LIST_SIZE(new_tmp);
#endif
			tmp=new_tmp;
			start_region_search(tmp,&tmpp);
			/* Left region is now handled -> skip to next */
			l_reg=first_of(&lp,&tmpp);
		}
	}
	free_gclist(r2);
	free_gclist(tmp);
	return a;
}

/*
 * Join operation
 */
struct GC_LIST *join(struct GC_LIST *l,int number)
{
	struct GC_LIST *a;
	struct GC_POINTER p1,p2;
	struct REGION r1,r2,prev_r1,prev_r2;
	int i;
	SGREPDATA(l);
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif
	
#ifdef DEBUG
	fprintf(stderr,"join called %d\n",number);
#endif
	assert(number>0);

	stats.join++;
	a=new_region_list(sgrep);
#ifdef OPTIMIZE_SORTS
	a->nested=l->nested;
#endif	
	if ( l->first==NULL )
	{
		/* This is an optimized chars node */
		to_chars(a,(l->chars+1)*number);
		return a;
	}
	
	/* List is smaller than join number, so return list is empty */
	if (LIST_SIZE(l)<number) return a;

#ifdef PROGRESS_REPORTS
	oper_name="join";
	prog=0;
	prog_start=LIST_SIZE(l);
#endif

	start_region_search(l,&p1);
	start_region_search(l,&p2);

	prev_r2.start=-1;	
	prev_r1.end=-1;	
	for (i=number;i>0;i--)
	{
		get_region(&p1,&r1);
		assert(r1.start!=-1);
	}
	while (r1.start!=-1)
	{
		get_region(&p2,&r2);
		if (r2.start==prev_r2.start)
		{
/*PK			if (prev.end<r2.end) */

			if (r1.end<=prev_r1.end)
			{
				list_set_sorted(a,NOT_SORTED);
			}
		}
		add_region(a,r2.start,r1.end);
		prev_r1=r1; 
		get_region(&p1,&r1);
		prev_r2=r2; 
	}
	if (list_get_sorted(a)!=START_SORTED)
	{ 
	  	remove_duplicates(a); /* There might be duplicates in a */
	};
	return a;	
}

/* 
 * Handles ordering which does _not_ produce nesting gc-lists. 
 * For example '"--" quote "--"' to catch SGML comments.
 */
struct GC_LIST *quote(struct GC_LIST *l,struct GC_LIST *r,int type)
{
	struct GC_POINTER lp,rp;
	struct GC_LIST *a;
	struct REGION r_reg,l_reg;              
	SGREPDATA(l);
#ifdef PROGRESS_REPORTS
	char *oper_name;
	int prog;int prog_start;
#endif

#ifdef DEBUG
	fprintf(stderr,"quote called\n");
#endif	
	stats.quote++;
	a=new_region_list(sgrep);

#ifdef PROGRESS_REPORTS
	prog_start=LIST_SIZE(r);
	prog=0;
	switch (type) {
	case L_QUOTE:
		oper_name="_quote";
		break;
	case R_QUOTE:
		oper_name="quote_";
		break;
	case LR_QUOTE:
		oper_name="_quote_";
		break;
	default:
		oper_name="quote";
	}
#endif
	start_region_search(r,&rp);
	start_region_search(l,&lp);
	get_region(&lp,&l_reg);
	get_region(&rp,&r_reg);

	/* If left or right region list was empty, we can return empty list */
	if (l_reg.start==-1 || r_reg.start==-1) return a;
	do {
		/* Skip until we find ending quote after start quote */
		while (l_reg.end>=r_reg.start && r_reg.start!=-1) {
			get_region(&rp,&r_reg);
		}
		if (r_reg.start>=0) {
			/* Add region using operation type */
			switch (type) {
			case QUOTE:
				add_region(a,l_reg.start, r_reg.end);
				break;
			case L_QUOTE:
				add_region(a,l_reg.end+1, r_reg.end);
				break;
			case R_QUOTE:
				add_region(a,l_reg.start, r_reg.start-1);
				break;
			case LR_QUOTE:
				/* No empty regions */
				if (l_reg.end+1<r_reg.start) {
					add_region(a,l_reg.end+1,r_reg.start-1);
				} 
				break;
			default:
			    abort();
			/* Skip until starting quote is after last ending
			   quote */
			}
			while (l_reg.start<=r_reg.end && l_reg.start!=-1) {
				get_region(&lp,&l_reg);
			}
		}
	} while ( l_reg.start!=-1 && r_reg.start!=-1);
	return a;
}

int find_first_starting_after(RegionList *l, int list_index, 
			      int region_index) {
    Region current;
    int jump=1;
    int ahead=0;
    while(jump>0) {
	region_at(l,list_index+ahead,&current);
	if (region_index<current.start) {
	    ahead=ahead+jump;
	    jump=jump*2;
	} else {
	    jump=jump/2;
	    ahead=ahead-jump;
	}
    }
    return list_index+ahead;
}


#define SGREP_INLINE /* FIXME __inline__ */
SGREP_INLINE int contains(Region *r1, Region *r2) {
    return 
	(r1->start<r2->start && r1->end>=r2->end) ||
	(r1->start<=r2->start && r1->end>r2->end);
}


RegionList *parenting(Evaluator *evaluator,RegionList *l, RegionList *r) {
    RegionList *result;
    Region parent,child;
    ListIterator parent_i,child_i;
    Region *stack=evaluator->tmp_stack;
    int sp=0; /* Stack pointer */
    SGREPDATA(l);


    /* Initialization */
    result=new_region_list(sgrep);
    list_set_sorted(result,NOT_SORTED);
    if (LIST_SIZE(r)>1) result->nested=1;
    start_region_search(r,&child_i);
    start_region_search(l,&parent_i);
    get_region(&parent_i,&parent);
    get_region(&child_i,&child);

    /* The parenting loop*/
    while(child.start!=-1) {

	/* This should be the most frequent case: parent candidate 
	 * ends before child has even started. Just simply skip to next
	 * parent */
	while(parent.end<child.start && parent.end!=-1) {
	    get_region(&parent_i,&parent);
	}

	/* Skip child candidates as long as
	 * 0. stack is empty
	 * 1. there is childs
	 * 2. child is not contained in parent
	 * 3. parent ends after child starts
	 */
	if (sp==0) {
	    while(child.start!=-1 && 
		  (!contains(&parent,&child)) &&
		  parent.end>child.start) {
		get_region(&child_i,&child);
	    }
	}

	/* No more childs --> nothing to do anymore */
	if (child.start==-1) break;
	
	if (parent.start!=-1 && contains(&parent,&child)) {
	    /* Now we know, that the child is contained in parent */

	    /* Now add the known good candidate to stack */
	    if (evaluator->tmp_stack_size==sp) {
		/* Allocate more space (previous_size*1.5) */
		evaluator->tmp_stack_size+=evaluator->tmp_stack_size/2;
		stack=(struct REGION *)sgrep_realloc(stack,
						     evaluator->tmp_stack_size*
						     sizeof(struct REGION));
		evaluator->tmp_stack=stack;
	    }
	    stack[sp++]=parent;
	    get_region(&parent_i,&parent);
	} else {
	    /* Now the innermost containing element is the topmost element
	     * in the stack. If stack is empty there is no containing element
	     */
	    if (sp>0) {
		assert(contains(&stack[sp-1],&child));
		add_region(result,stack[sp-1].start,stack[sp-1].end);
		/* This child is now handled. Next child */
		get_region(&child_i,&child);

		/* Now remove the regions which do not contain the new child
		 * from stack */ 
		if (child.start!=-1) {
		    while(sp>0 && stack[sp-1].end<child.end) {
			assert(stack[sp-1].start<=child.start);
			sp--;		    
		    }
		}
	    }	    
	}
    }
    /* The result list is not sorted and might contain duplicates */
    remove_duplicates(result);
    return result;
}
