/*
	System: Structured text retrieval tool sgrep.
	Module: parser.c
	Author: Pekka Kilpelinen & Jani Jaakkola
	Description: Parses expression string ( which has been gathered in
	             main.c and preprocessed in preprocess.c )
	             Returns operator tree.
	             Used through fuction parse_tree()
	Version history: Original version February 1995 by JJ & PK
	Copyright: University of Helsinki, Dept. of Computer Science
		   Distributed under GNU General Public Lisence
		   See file COPYING for details
*/
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "sgrep.h"


/* Maximum length of one phrase. */
#define MAX_PHRASE_LENGTH 512
/* Maximum length of a reserved word. Could also be smaller */
#define MAX_R_WORD 20
#define ELLENGTH 79
#define SHOWR 5

/* 
 * Tokens of sgrep query language
 */
enum Token { 
    /* Tokens for reserved words */
    W_IN,W_NOT,W_CONTAINING,W_OR,
    W_ORDERED,W_L_ORDERED,W_R_ORDERED,W_LR_ORDERED,
    W_EXTRACTING,
    W_QUOTE,W_L_QUOTE,W_R_QUOTE,W_LR_QUOTE,
    W_EQUAL,		/* PK Febr 12 '96 */
    W_PARENTING,W_CHILDRENING,
    W_OUTER,W_INNER,W_CONCAT,W_START,W_LAST,W_CHARS,
    W_JOIN,
    W_FILE,W_PROLOG,
    /* Phrase types */
    W_ELEMENTS,
    W_STRING, W_REGEX,
    W_DOCTYPE, W_DOCTYPE_PID, W_DOCTYPE_SID,
    W_PI, W_ATTRIBUTE, W_ATTVALUE,
    W_STAG, W_ETAG, W_COMMENT, W_COMMENT_WORD,
    W_WORD, W_CDATA,
    W_ENTITY, W_ENTITY_DECLARATION, W_ENTITY_LITERAL,
    W_ENTITY_PID, W_ENTITY_SID, W_ENTITY_NDATA,
    W_RAW,
    R_WORDS, 
    /* And the rest */
    W_LPAREN,W_RPAREN,W_LBRACK,W_RBRACK,W_COMMA,W_PHRASE,W_NUMBER,W_END,
/* In sgrep, parse error while scanning is just a different kind of token */
    W_PARSE_ERROR,
    W_END_TOKENS
};


/*
 * Parser object
 */
typedef struct {
    SgrepData *sgrep;
    /* The command string and a index to it */
    char *expr_str;
    int expr_ind;

    /* Current character, token and string of current string token */
    int ch;
    enum Token token;
    SgrepString *string_token;
    
    /* This points to first phrase in the phrase list */
    struct PHRASE_NODE *first_phrase;

    /* for telling where errors occurred */
    int line,column;
    int errcol,errind;

    /* Scanners state */
    int scanner_state;

} Parser;



void real_parse_error(Parser *parser,char *);
int get_next_char(Parser *parser);
enum Token next_token(Parser *parser);

struct TREE_NODE *create_tree_node(Parser *parser,int oper);
struct TREE_NODE *parse_reg_expr(Parser *parser);
struct TREE_NODE *parse_basic_expr(Parser *parser);
struct TREE_NODE *parse_oper_expr(Parser *parser,struct TREE_NODE *);
struct TREE_NODE *parse_phrase(Parser *,const char *);
struct TREE_NODE *parse_cons_list(Parser *parser);
struct TREE_NODE *new_string_phrase(Parser *, SgrepString *, 
					   const char *);
#ifdef DEBUG
void print_tree(struct TREE_NODE *);
#endif

#define parse_error(E) \
  do { real_parse_error(parser,(E));return NULL; } while (0)
#define token (parser->token)
#define NEXT_TOKEN \
    do { token=next_token(parser); if (token==W_PARSE_ERROR) return NULL; } while (0);


/*
 * When looking for reserved words they are matched to these strings 
 */
struct {
    const char *word;
    enum Token t;
} reserved_words[]=
{ 
    {"in",W_IN},
    {"not",W_NOT},
    {"containing",W_CONTAINING},
    {"or",W_OR},
    {"..",W_ORDERED},
    {"_.",W_L_ORDERED},
    {"._",W_R_ORDERED},
    {"__",W_LR_ORDERED},
    {"extracting", W_EXTRACTING},
    {"quote",W_QUOTE},
    {"_quote",W_L_QUOTE},
    {"quote_",W_R_QUOTE},
    {"_quote_",W_LR_QUOTE },
    {"equal",W_EQUAL},	 /* PK Febr 12 '96 */
    {"parenting",W_PARENTING},
    {"childrening",W_CHILDRENING},
    {"outer",W_OUTER},
    {"inner",W_INNER},
    {"concat",W_CONCAT},
    /* Constant lists */
    {"start",W_START},
    {"end",W_LAST},
    {"chars",W_CHARS},
    {"join",W_JOIN},
    {"file",W_FILE},
    /* Different phrases */
    {"string", W_STRING},
    {"regex",W_REGEX},
    {"pi", W_PI},
    {"attribute", W_ATTRIBUTE},
    {"attvalue",W_ATTVALUE},
    {"stag", W_STAG},
    {"etag", W_ETAG},
    {"comments", W_COMMENT},
    {"elements", W_ELEMENTS},
    {"comment_word",W_COMMENT_WORD},
    {"word", W_WORD},
    {"cdata",W_CDATA},
    {"entity",W_ENTITY},

    /* Stuff from DTD */
    {"prologs", W_PROLOG},
    {"doctype", W_DOCTYPE},
    {"doctype_pid", W_DOCTYPE_PID},
    {"doctype_sid",W_DOCTYPE_SID},
    {"entity_declaration",W_ENTITY_DECLARATION },
    {"entity_literal",W_ENTITY_LITERAL },
    {"entity_pid",W_ENTITY_PID },
    {"entity_sid",W_ENTITY_SID },
    {"entity_ndata",W_ENTITY_NDATA },
    /* And the raw phrase */
    {"raw", W_RAW },
    {NULL,W_END_TOKENS}
};

/*
 * Gives user readable operation name from given opernum
 */
const char *give_oper_name(int oper)
{
	switch (oper) {
	case IN:		return "in";break;
	case NOT_IN:		return "not in";break;
	case CONTAINING:	return "containing";break;
	case NOT_CONTAINING:	return "not containing";break;
	case EQUAL:		return "equal";break; /* PK Febr'95 */
	case NOT_EQUAL:		return "not equal";break; /* PK Febr'95 */
	case OR:		return "or";break;
	case ORDERED:		return "..";break;
	case L_ORDERED:		return "_.";break;
	case R_ORDERED:		return "._";break;
	case LR_ORDERED:	return "__";break;
	case EXTRACTING:	return "extracting";break;
	case OUTER:		return "outer";break;
	case INNER:		return "inner";break;
	case CONCAT:		return "concat";break;
	case JOIN:		return "join";break;
	case PHRASE:		return "phrase";break;
	case INVALID:		return "invalid";break;
	default:		return "unknown";break;
	}
}


/*
 * Gets next character from input expression. Counts also lines and columns
 * return character must be unsigned to keep sgrep is 8-bit clean
 */

int get_next_char(Parser *parser)
{
	switch (parser->expr_str[parser->expr_ind]) {
		case '\n':
			parser->column=-1;
			parser->line++;
			break;
		case '\t':
			parser->expr_str[parser->expr_ind]=' ';
			break;
	}
	parser->column++;
	return parser->ch=((unsigned char *)parser->expr_str)[parser->expr_ind++];	
}
	
/*
 * Scans next token from input string. If token was a phrase sets **phrase 
 */
enum Token next_token(Parser *parser)
{
	char str[MAX_PHRASE_LENGTH];
	int i=0,j=0;
	int start;
	int onum=0;
	int ch;
	SgrepString **phrase;
	SGREPDATA(parser);

/* if we already have next character in ch then ch!=-1. Otherwise we have to
   get one */
	ch=parser->ch;
	phrase=&parser->string_token;
	if (ch==-1) ch=get_next_char(parser);
 
/* These show the place where possible error started */
	parser->errcol=parser->column;
	parser->errind=parser->expr_ind;

/* Finite automaton for recognizing next word */	
	do {
	switch (parser->scanner_state) {
	
	/* This is the start state */
	case 0: 
		/* skip the whitespace */
		while (isspace(ch)) ch=get_next_char(parser);

		/* These show the place where possible error started */
		parser->errcol=parser->column;
		parser->errind=parser->expr_ind;
		
		switch(ch) {
		case 0:
		        
			return W_END;
		case '(':
			parser->ch=-1;
			return W_LPAREN;
		case ')': 
			parser->ch=-1;
			return W_RPAREN;
		case '[':
			parser->ch=-1;
			return W_LBRACK;
		case ']':
			parser->ch=-1;
			return W_RBRACK;
		case ',':
			parser->ch=-1;
			return W_COMMA;
		case '\"': 
			i=0;
			parser->scanner_state=1;
			break;
		case '#':
			parser->scanner_state=4;
			break;
		}
		if ( isalpha(ch) || ch=='.' || ch=='_' )
		{
			i=1;
			str[0]=tolower(ch);
			parser->scanner_state=2;
			start=parser->column;
		}
		if (ch>='0' && ch<='9')
		{
			i=1;
			str[0]=ch;
			start=parser->column;
			parser->scanner_state=5;
		}
		if ( parser->scanner_state==0 )
		{
#ifdef DEBUG
			fprintf(stderr,"Character %c ascii(%d)\n",ch,ch);
#endif
			real_parse_error(parser,"Invalid character");
			return W_PARSE_ERROR;
		}
		break;
		
	/* This state reads a phrase string */
	case 1:
	    if (ch==0 || ch=='\n') 
 {
		real_parse_error(parser,"Unterminated phrase string");
		return W_PARSE_ERROR;
	    }
	    if ( ch<' ' ) {
		real_parse_error(parser,"Unprintable character in phrase");
		return W_PARSE_ERROR;
	    }
	    if (ch=='\"') 
	    {
		if ( i==0 )
 {
		    real_parse_error(parser,"Empty phrase");
		}
		*phrase=init_string(sgrep,i,str);
		parser->scanner_state=0;
		parser->ch=-1;
		return W_PHRASE;
	    } 	
	    if ( i==MAX_PHRASE_LENGTH )
	    {
		sgrep_error(sgrep,"%s ( > %d ) %s %d\n%s\n%s\n",
	"Phrase length exceeds",MAX_PHRASE_LENGTH,"on line",parser->line,
	"Either you have forgotten the quote  at the end of phrase or",
	"you have to recompile with greater MAX_PHRASE_LENGTH.");
		real_parse_error(parser,"");
		return W_PARSE_ERROR;
	    }
	    if (ch=='\\' ) parser->scanner_state=3;
	    if ( parser->scanner_state==1 ) str[i++]=ch;
	    break;
	    
	    /* This state reads a reserved word (operator) */
	case 2:
	    if ( ch==0 || ( !isalpha(ch) && ch!='.' && ch!='_' ) )
	    {
		parser->scanner_state=0;
		str[i]=0;
		for(j=0;reserved_words[j].word;j++)
		    if (strcmp(reserved_words[j].word,str)==0) 
			return reserved_words[j].t;
		real_parse_error(parser,"Unknown word");
		return W_PARSE_ERROR;
	    }
	    if (i==MAX_R_WORD)
 {
		real_parse_error(parser,"Unknown word");
		return W_PARSE_ERROR;
	    }
	    str[i++]=tolower(ch);
	    break;
	    
	/* This state handles escape sequences */
	case 3:
		if ( ch==0 ) {		    
		    real_parse_error(parser,"Unterminated phrase");
		    return W_PARSE_ERROR;
		}
		switch (ch) {
		case 't': ch='\t';break;
		case 'n': ch='\n';break;
		case 'r': ch='\r';break;
		case 'f': ch='\f';break;
		case 'b': ch='\b';break;
		case '\\': ch='\\';break;
		case '\"': ch='\"';break;
		case '0':
		case '1':
		case '2':
		case '3':
			/* Octal number requires new state */
			parser->scanner_state=6;
			onum=64*(ch-'0');
			break;
		default:
			parser->errcol=parser->column;
			parser->errind=parser->expr_ind;
			real_parse_error(parser,"Unknown escape sequence");
			return W_PARSE_ERROR;
			break;
		}
                /* always true except for octal number */
		if (parser->scanner_state==3) 
		{
			str[i++]=ch;
			parser->scanner_state=1;
		}
		break;
	/* This state skips comments */
	case 4:
		if (ch==0)
		{
			parser->scanner_state=0;
			return W_END;
		}
		if (ch=='\n') parser->scanner_state=0;
		break;
	/* We read a number */
	case 5:
		if ( ch<'0'|| ch>'9' )
		{
			parser->scanner_state=0;
			str[i]=0;
			*phrase=init_string(sgrep,i,str);
			return W_NUMBER;
		}
		if (i==9)
 {
			real_parse_error(parser,"Too big number");
			return W_PARSE_ERROR;
		}
		str[i++]=ch;
		break;
	/* Read octal number #2 */
	case 6:
		if ( ch<'0' || ch>'7')
		{
			parser->errcol=parser->column;
			parser->errind=parser->expr_ind;
			real_parse_error(parser,"Octal number expected");
			return W_PARSE_ERROR;
		}
		onum+=8*(ch-'0');
		parser->scanner_state=7;
		break;
	/* Read octal number #3 */
	case 7:
		if ( ch<'0' || ch>'7')
		{
			parser->errcol=parser->column;
			parser->errind=parser->expr_ind;
			real_parse_error(parser,"Octal number expected");
			return W_PARSE_ERROR;
		}
		onum+=ch-'0';
		if (onum==0)
		{
                        parser->errcol=parser->column-3;
                        parser->errind=parser->expr_ind-3;                        
			real_parse_error(parser,"\\000 in phrase does'nt work.\n");
			return W_PARSE_ERROR;

		}
		str[i++]=onum;
		parser->scanner_state=1;
		break;
	}
	
	ch=get_next_char(parser);
	} while (TRUE);

}

/* 
 * Shows a given error message & where it occurred 
 */
void real_parse_error(Parser *parser, char *error)
{
    	char erlin[ELLENGTH+1];
	int i;
	SGREPDATA(parser);
	
	if (parser->errcol-ELLENGTH+SHOWR>0)
		parser->errind-=ELLENGTH-SHOWR;
	else parser->errind-=parser->errcol;
	for(i=0;i<ELLENGTH && parser->expr_str[i+parser->errind] && parser->expr_str[i+parser->errind]!='\n';i++) 
		erlin[i]=parser->expr_str[i+parser->errind];
	erlin[i]=0;
	sgrep_error(parser->sgrep,"Parse error on line %d column %d :\n\t%s\n%s\n",
		parser->line,parser->errcol,error,erlin); 
	if ( parser->errcol>ELLENGTH-SHOWR ) parser->errcol=ELLENGTH-SHOWR; 
        for (i=0;i<parser->errcol-1;i++) sgrep_error(sgrep," "); 
        sgrep_error(sgrep,"^\n");
}

/*
 * Creates and initializez an operator node to parse tree
 */
struct TREE_NODE *create_tree_node(Parser *parser, int oper)
{
	struct TREE_NODE *n;
	SgrepData *sgrep=parser->sgrep;

	n=sgrep_new(struct TREE_NODE);
	n->left=NULL;
	n->right=NULL;
	n->parent=NULL;
	n->oper=oper;
	n->number=-1;
	n->leaf=NULL;
	n->label_left=LABEL_NOTKNOWN;
	n->label_right=LABEL_NOTKNOWN;
	n->refcount=0;
	n->GC_list=NULL;
	return n;
}
	
/* 
 * Creates and initializes a leaf node to parse tree 
 */
struct TREE_NODE *create_leaf_node(Parser *parser, int oper, 
				   SgrepString *phrase, int phrase_label)
{
    struct TREE_NODE *n;
    SgrepData *sgrep=parser->sgrep;

    /* Create a leaf node */
    n=create_tree_node(parser,oper);
    n->label_left=phrase_label;
    /* Now create a phrase node for the phrase */
    n->leaf=sgrep_new(struct PHRASE_NODE);
    n->leaf->phrase=phrase;
    n->leaf->GC_list=NULL;
    n->GC_list=NULL;
    return n;
}

struct TREE_NODE *new_string_phrase(Parser *parser,
					   SgrepString *phrase, 
					   const char *typestr) {
    struct TREE_NODE *n;
    SGREPDATA(parser);

    if (phrase==NULL) {
	phrase=init_string(sgrep,strlen(typestr),typestr);
    } else {
	push_front(phrase,typestr);
    }
    n=create_leaf_node(parser,PHRASE,phrase,LABEL_PHRASE);
    /* Put the new phrase node to phrase list */
    n->leaf->next=parser->first_phrase;
    parser->first_phrase=n->leaf;
    return n;
}


/* 
 * Here starts recursive parser 
 */

struct TREE_NODE *parse_phrase(Parser *parser, const char *typestr) {
    struct TREE_NODE *n;
    NEXT_TOKEN;
    if (token!=W_LPAREN) parse_error("Expecting '('");
    NEXT_TOKEN;
    if (token!=W_PHRASE) parse_error("Expecting phrase string");
    n=new_string_phrase(parser,parser->string_token,typestr);
#if DEBUG
    fprintf(stderr,"parse_phrase '%s'\n",n->leaf->phrase->s);
#endif
    NEXT_TOKEN;
    if (token!=W_RPAREN) parse_error("Expecting ')'");
    NEXT_TOKEN;
    return n;
}


/* production basic_expr->constant_list */
struct TREE_NODE *parse_cons_list(Parser *parser)
{
    int s,e,ps,pe;
    char *cons_err="invalid constant region list";
    struct TREE_NODE *n;
    struct GC_LIST *c_list;
    SGREPDATA(parser);
    
    n=create_leaf_node(parser,PHRASE,NULL,LABEL_CONS);
    NEXT_TOKEN;
    c_list=n->leaf->GC_list=new_region_list(sgrep);
    
    ps=-1;pe=-1;
    while (token!=W_RBRACK) /* We can fall out here immediately,
			       which means we have empty list */
    {
	if (token!=W_LPAREN)
	    parse_error(cons_err);
	NEXT_TOKEN;
	if (token!=W_NUMBER)
	    parse_error(cons_err);
	s=atoi((char *)(parser->string_token->s));
	NEXT_TOKEN;
	if (token!=W_COMMA)
	    parse_error(cons_err);
	NEXT_TOKEN;
	if (token!=W_NUMBER)
	    parse_error(cons_err);
	e=atoi((char *)(parser->string_token->s));
	NEXT_TOKEN;
	if (token!=W_RPAREN)
	    parse_error(cons_err);
	if (e<s)
	    parse_error("region end point must be greater than start point");
	NEXT_TOKEN;
	if (s<ps || (s==ps && e<=pe) )
	    parse_error("constant gc list must be sorted");
	if (e<=pe || s==ps)
	{
	    /* nesting detected */
	    c_list->nested=TRUE;
	}
	add_region(c_list,s,e);
	ps=s;pe=e;
    }
    parser->sgrep->statistics.constant_lists++;
    NEXT_TOKEN;
    return n;
}

/* Which kind of basic expressions */			
struct TREE_NODE *parse_basic_expr(Parser *parser) {
    struct TREE_NODE *n;
    
    switch (token) {
    case W_PHRASE:
	/* A simple phrase */
	n=new_string_phrase(parser,parser->string_token, "n");
	NEXT_TOKEN;
	return n;
    case W_START:
	/* reserved word 'start' was found */
	n=create_leaf_node(parser,PHRASE,NULL,LABEL_START);
	/* we use already created list (see create_constant_lists() )*/
	n->leaf->GC_list=parser->sgrep->start_list;
	NEXT_TOKEN;
	return n;
    case W_LAST:
	/* reserved word 'end' was found */
	n=create_leaf_node(parser,PHRASE,NULL,LABEL_END);
	/* we use already created list (see create_constant_lists() )*/
	n->leaf->GC_list=parser->sgrep->end_list;
	NEXT_TOKEN;
	return n;
    case W_LBRACK:
	/* We got a constant region list */
	return parse_cons_list(parser);	
    case W_CHARS:
	/* reserved wors 'chars' was found */
	n=create_leaf_node(parser,PHRASE,NULL,LABEL_CHARS);
	/* we use already created list (see create_constant_lists() )*/
	n->leaf->GC_list=parser->sgrep->chars_list;
	NEXT_TOKEN;
	return n;
    case W_LPAREN:
	NEXT_TOKEN;
	n=parse_reg_expr(parser);
	if (n==NULL) return NULL;
	if (token!=W_RPAREN)
	    parse_error(") expected");
	NEXT_TOKEN;
	return n;
    case W_OUTER:
    case W_INNER:
    case W_CONCAT:
	/* So we have function. Let's create a tree node for it */
	n=create_tree_node(parser, 
			   (token==W_OUTER) ? OUTER:
			   ((token==W_INNER) ? INNER:CONCAT) ); 
	/* Let's parse the parameter */
	NEXT_TOKEN;
	if (token!=W_LPAREN)
	    parse_error("( expected");
	NEXT_TOKEN;
	n->left=parse_reg_expr(parser);
	if (n->left==NULL) return NULL;
	n->right=NULL; /* Function has only one parameter */
	if (token!=W_RPAREN)
	    parse_error(") expected");
	NEXT_TOKEN;
	return n;
    case W_JOIN:
	/* So we have join function. Let's create a tree node for it */
	n=create_tree_node(parser,JOIN);
	/* Let's parse the parameters */
	NEXT_TOKEN;
	if (token!=W_LPAREN)
	    parse_error("( expected");
	NEXT_TOKEN;
	if (token!=W_NUMBER)
	    parse_error("join needs a number: join(NUMBER:expression)");
	n->number=atoi((char *)parser->string_token->s);
	if (n->number<1)
	    parse_error("join number must be >= 1");
	NEXT_TOKEN;
	if (token!=W_COMMA)
	    parse_error("join expected comma: join(NUMBER,expression)");
	NEXT_TOKEN;
	n->left=parse_reg_expr(parser);
	if (n->left==NULL) return NULL;
	n->right=NULL;
	if (token!=W_RPAREN)
	    parse_error(") expected");
	NEXT_TOKEN;
	return n;
	/* phrases */
    case W_COMMENT: 
	n=new_string_phrase(parser,NULL,"-");
	NEXT_TOKEN;
	return n;
    case W_ELEMENTS:
	n=new_string_phrase(parser,NULL,"@elements");
	NEXT_TOKEN;
	return n;
    case W_PROLOG:
	n=new_string_phrase(parser,NULL,"d!");
	NEXT_TOKEN;
	return n;
    case W_CDATA:
	n=new_string_phrase(parser,NULL,"[CDATA");
	NEXT_TOKEN;
	return n;
	/* phrases with argument */
    case W_RAW:
	return parse_phrase(parser,"");
    case W_FILE:
	return parse_phrase(parser,"f");
    case W_STRING: 
	return parse_phrase(parser,"n");
    case W_REGEX:
	return parse_phrase(parser,"r");
    case W_DOCTYPE:
	return parse_phrase(parser, "dn");
    case W_DOCTYPE_PID: 
	return parse_phrase(parser, "dp");
    case W_DOCTYPE_SID:
	return parse_phrase(parser,"ds");
    case W_ENTITY_DECLARATION:
	return parse_phrase(parser,"!ed");
    case W_ENTITY_LITERAL:
	return parse_phrase(parser,"!el");
    case W_ENTITY_PID:
	return parse_phrase(parser,"!ep");
    case W_ENTITY_SID:
	return parse_phrase(parser,"!es");
    case W_ENTITY_NDATA:
	return parse_phrase(parser,"!en");
    case W_PI:
	return parse_phrase(parser,"?");	    
    case W_ATTRIBUTE:
	return parse_phrase(parser,"a");
    case W_ATTVALUE:
	return parse_phrase(parser,"v");	    
    case W_STAG: 
	return parse_phrase(parser,"s");
    case W_ETAG:
	return parse_phrase(parser,"e");
    case W_COMMENT_WORD:
	return parse_phrase(parser,"c");	    
    case W_WORD: 
	return parse_phrase(parser,"w");
    case W_ENTITY:
	return parse_phrase(parser,"&");

	
    default:
	parse_error("Basic expression expected\n");
    }

    /* Dummy return for keeping c++ compilers quiet */
    return NULL;
}	       

/* Which kind region_expression */
struct TREE_NODE *parse_reg_expr(Parser *parser)
{
	struct TREE_NODE *left;
	if (token==W_END)
		parse_error("Unexpected end of expression");

	left=parse_basic_expr(parser);
	if (left==NULL) return NULL;
	if ( token==W_END || token==W_RPAREN ) return left;
	return parse_oper_expr(parser,left);
}	

/* Which kind of operator expression */
struct TREE_NODE *parse_oper_expr(Parser *parser,struct TREE_NODE *left)
{
	struct TREE_NODE *o=NULL;
	switch (token)
	{
	case W_IN:		
		o=create_tree_node(parser,IN);
		break;
	case W_CONTAINING:	
		o=create_tree_node(parser,CONTAINING);
		break;
/* PK Febr 95 */
	case W_EQUAL:	
		o=create_tree_node(parser,EQUAL);
		break;
/* PK Febr 95 */
	case W_PARENTING:
	        o=create_tree_node(parser,PARENTING);
                break;
	case W_CHILDRENING:
	        o=create_tree_node(parser,CHILDRENING);
		break;
	case W_OR: 		
		o=create_tree_node(parser,OR);
		break;
	case W_ORDERED: 
		o=create_tree_node(parser,ORDERED);
		break;
	case W_L_ORDERED:
		o=create_tree_node(parser,L_ORDERED);
		break;
	case W_R_ORDERED:
		o=create_tree_node(parser,R_ORDERED);
		break;
	case W_LR_ORDERED:
		o=create_tree_node(parser,LR_ORDERED);
		break;
	case W_EXTRACTING:
		o=create_tree_node(parser,EXTRACTING);
		break;
	case W_QUOTE:
		o=create_tree_node(parser,QUOTE);
		break;
	case W_R_QUOTE:
		o=create_tree_node(parser,R_QUOTE);
		break;
	case W_L_QUOTE:
		o=create_tree_node(parser,L_QUOTE);
		break;
	case W_LR_QUOTE:
		o=create_tree_node(parser,LR_QUOTE);
		break;
	case W_NOT:
		NEXT_TOKEN;
		if (token==W_CONTAINING) o=create_tree_node(parser,NOT_CONTAINING);
		else if (token==W_IN) o=create_tree_node(parser,NOT_IN);
		else if (token==W_EQUAL) o=create_tree_node(parser,NOT_EQUAL);
		else parse_error("'not' must be followed by 'in', 'containing' or 'equal'");
		break;
	default:
		parse_error("Operator expected");
	}
	NEXT_TOKEN;
	o->right=parse_basic_expr(parser);
	if (o->right==NULL) return NULL;
	o->left=left;
	if (token==W_END || token==W_RPAREN) return o;
	return parse_oper_expr(parser,o);
}		

/* End of recursive parser */

/*
 * Parses the given command string. Returns root node of the parse tree.
 * phrase_list returns pointer to phrase list 
 */
struct TREE_NODE *parse_string(SgrepData *sgrep,const char *str, 
	struct PHRASE_NODE **phrase_list)
{
	struct TREE_NODE *root;
	Parser parser_l;
	Parser *parser=&parser_l;
	
	parser->sgrep=sgrep;
	parser->line=1;
	parser->column=0;
	parser->expr_str=(char *) str;
	parser->expr_ind=0;
	parser->ch=-1;
	parser->first_phrase=NULL;
	parser->scanner_state=0;

	NEXT_TOKEN;
	root=parse_reg_expr(parser);

	if (token==W_RPAREN)
	{
		real_parse_error(parser,"Too many )'s");
	}
	if (token!=W_END)
	{
	    return NULL;
	}

#ifdef DEBUG
	print_tree(root);
	fprintf(stderr,"\n");
#endif	
	*phrase_list=parser->first_phrase;
	return root;	
}


void free_parse_tree(SgrepData *sgrep, struct TREE_NODE *root) {
    assert(root->refcount!=0);
    if (root->refcount==-1) {
	assert(root->oper==PHRASE && root->leaf && root->leaf->phrase==NULL);
	root->refcount=0;
	sgrep_free(root->leaf);
	sgrep_free(root);
    } else {
	if (root->left) free_parse_tree(sgrep,root->left);
	if (root->right) free_parse_tree(sgrep,root->right);
	root->refcount--;
	if (root->refcount==0) {
	    if (root->oper==PHRASE) {
		assert(root->leaf);
		delete_string(root->leaf->phrase);
		sgrep_free(root->leaf);
	    }
	    sgrep_free(root);
	}
    }
}
	
	
#ifdef DEBUG
void print_tree(struct TREE_NODE *root)
{
	if ( root==NULL )
	{
		fprintf(stderr,"\nprint_tree: got NULL node\n");
		exit(3);
	}
	if ( root->oper==PHRASE )
	{
		fprintf(stderr," \"%s\"",root->leaf->phrase->s);
		return;
	}
	if (root->oper<0 || root->oper>R_WORDS)
	{
		printf("\nprint tree: got invalid oper (%d)\n",root->oper);
		exit(3);
	}
	
	if (root->right!=NULL)
	{
		fprintf(stderr," (");
		print_tree(root->left);
	
		fprintf(stderr," %s",r_word_str[root->oper]);
	
		print_tree(root->right);
		fprintf(stderr," )");
	} else
	{
		fprintf(stderr," %s(",r_word_str[root->oper]);
		print_tree(root->left);
		fprintf(stderr," )");
	}
}
#endif
